# wireless-sensor-mqtt - MQTT client reporting measurements of FT017TH wireless thermo/hygrometers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import wireless_sensor_mqtt._homeassistant

# pylint: disable=protected-access


@pytest.mark.parametrize(
    ("node_id", "valid"),
    [
        ("raspberrypi", True),
        ("da-sh", True),
        ("under_score", True),
        ('" or ""="', False),
        ("", False),
    ],
)
def test_validate_node_id(node_id, valid):
    assert wireless_sensor_mqtt._homeassistant.validate_node_id(node_id) == valid



# wireless-sensor-mqtt - MQTT client reporting measurements of FT017TH wireless thermo/hygrometers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import json
import unittest.mock

import pytest

import wireless_sensor
import wireless_sensor_mqtt


# pylint: disable=protected-access


@pytest.mark.parametrize(
    "mqtt_topic_prefix", ["wireless-sensor/FT017TH", "living-room/ft017th"]
)
@pytest.mark.parametrize("homeassistant_discovery_prefix", ["homeassistant"])
@pytest.mark.parametrize(
    "homeassistant_node_id", ["ft017th-living-room", "bed-room-sensor"]
)
def test__publish_homeassistant_discovery_config(
    mqtt_topic_prefix, homeassistant_discovery_prefix, homeassistant_node_id
):
    mqtt_client_mock = unittest.mock.MagicMock()
    wireless_sensor_mqtt._publish_homeassistant_discovery_config(
        mqtt_client=mqtt_client_mock,
        homeassistant_discovery_prefix=homeassistant_discovery_prefix,
        homeassistant_node_id=homeassistant_node_id,
        temperature_topic=mqtt_topic_prefix + "/temp",
        humidity_topic=mqtt_topic_prefix + "/rel-humidity",
    )
    publish_calls_args = mqtt_client_mock.publish.call_args_list
    assert len(publish_calls_args) == 2
    for call_args in publish_calls_args:
        assert not call_args[0]  # positional args
        assert set(call_args[1].keys()) == {"topic", "payload", "retain"}
        assert call_args[1]["retain"] is True
    assert publish_calls_args[0][1][
        "topic"
    ] == "homeassistant/sensor/{}/temperature-degrees-celsius/config".format(
        homeassistant_node_id
    )
    assert publish_calls_args[1][1][
        "topic"
    ] == "homeassistant/sensor/{}/relative-humidity-percent/config".format(
        homeassistant_node_id
    )
    device_attrs = {
        "identifiers": ["FT017TH/" + homeassistant_node_id],
        "model": "FT017TH",
    }
    assert json.loads(publish_calls_args[0][1]["payload"]) == {
        "unique_id": "fphammerle/wireless-sensor-mqtt/FT017TH/{}/{}".format(
            homeassistant_node_id, "temperature-degrees-celsius"
        ),
        "name": "{} temperature".format(homeassistant_node_id),
        "state_topic": mqtt_topic_prefix + "/temp",
        "device_class": "temperature",
        "unit_of_measurement": "°C",
        "expire_after": 600,
        "device": device_attrs,
    }
    assert json.loads(publish_calls_args[1][1]["payload"]) == {
        "unique_id": "fphammerle/wireless-sensor-mqtt/FT017TH/{}/relative-humidity-percent".format(
            homeassistant_node_id
        ),
        "name": "{} relative humidity".format(homeassistant_node_id),
        "state_topic": mqtt_topic_prefix + "/rel-humidity",
        "device_class": "humidity",
        "unit_of_measurement": "%",
        "expire_after": 600,
        "device": device_attrs,
    }


@pytest.mark.parametrize("mqtt_host", ["mqtt-broker.local"])
@pytest.mark.parametrize("mqtt_port", [1234])
@pytest.mark.parametrize("mqtt_disable_tls", [True, False])
@pytest.mark.parametrize(
    ("mqtt_username", "mqtt_password"),
    [(None, None), ("user", None), ("username", "secret")],
)
@pytest.mark.parametrize(
    "mqtt_topic_prefix", ["wireless-sensor/FT017TH", "living-room/ft017th"]
)
@pytest.mark.parametrize("homeassistant_discovery_prefix", ["homeassistant"])
@pytest.mark.parametrize("homeassistant_node_id", ["ft017th-living-room"])
@pytest.mark.parametrize("unlock_spi_device", [True, False])
def test__run(
    mqtt_host,
    mqtt_port,
    mqtt_disable_tls,
    mqtt_username,
    mqtt_password,
    mqtt_topic_prefix,
    homeassistant_discovery_prefix,
    homeassistant_node_id,
    unlock_spi_device,
):
    # pylint: disable=too-many-arguments
    with unittest.mock.patch(
        "wireless_sensor.FT017TH.__init__", return_value=None
    ) as sensor_init_mock, unittest.mock.patch(
        "wireless_sensor.FT017TH.receive",
        side_effect=[
            [
                wireless_sensor.Measurement(
                    decoding_timestamp=datetime.datetime(2020, 12, 7, 18, 5, 1),
                    temperature_degrees_celsius=23.1234567,
                    relative_humidity=0.501234567,
                ),
                wireless_sensor.Measurement(
                    decoding_timestamp=datetime.datetime(2020, 12, 7, 18, 6, 19),
                    temperature_degrees_celsius=24.1234567,
                    relative_humidity=0.401234567,
                ),
            ]
        ],
    ):
        mqtt_client_mock = unittest.mock.MagicMock()
        with unittest.mock.patch(
            "wireless_sensor_mqtt._init_mqtt_client"
        ) as init_mqtt_client_mock, unittest.mock.patch(
            "wireless_sensor_mqtt._publish_homeassistant_discovery_config"
        ) as hass_config_mock:
            init_mqtt_client_mock.return_value = mqtt_client_mock
            wireless_sensor_mqtt._run(
                mqtt_host=mqtt_host,
                mqtt_port=mqtt_port,
                mqtt_disable_tls=mqtt_disable_tls,
                mqtt_username=mqtt_username,
                mqtt_password=mqtt_password,
                mqtt_topic_prefix=mqtt_topic_prefix,
                homeassistant_discovery_prefix=homeassistant_discovery_prefix,
                homeassistant_node_id=homeassistant_node_id,
                mock_measurements=False,
                unlock_spi_device=unlock_spi_device,
            )
    init_mqtt_client_mock.assert_called_once_with(
        host=mqtt_host,
        port=mqtt_port,
        disable_tls=mqtt_disable_tls,
        username=mqtt_username,
        password=mqtt_password,
    )
    sensor_init_mock.assert_called_once_with(unlock_spi_device=unlock_spi_device)
    hass_config_mock.assert_called_once_with(
        mqtt_client=mqtt_client_mock,
        homeassistant_discovery_prefix=homeassistant_discovery_prefix,
        homeassistant_node_id=homeassistant_node_id,
        temperature_topic=mqtt_topic_prefix + "/temperature-degrees-celsius",
        humidity_topic=mqtt_topic_prefix + "/relative-humidity-percent",
    )
    publish_calls_args = mqtt_client_mock.publish.call_args_list
    assert len(publish_calls_args) == 2 * 2
    for call_args in publish_calls_args:
        assert not call_args[0]  # positional args
        assert set(call_args[1].keys()) == {"topic", "payload", "retain"}
        assert call_args[1]["retain"] is False
    assert (
        publish_calls_args[0][1]["topic"]
        == publish_calls_args[2][1]["topic"]
        == mqtt_topic_prefix + "/temperature-degrees-celsius"
    )
    assert (
        publish_calls_args[1][1]["topic"]
        == publish_calls_args[3][1]["topic"]
        == mqtt_topic_prefix + "/relative-humidity-percent"
    )
    assert publish_calls_args[0][1]["payload"] == "23.12"
    assert publish_calls_args[1][1]["payload"] == "50.12"
    assert publish_calls_args[2][1]["payload"] == "24.12"
    assert publish_calls_args[3][1]["payload"] == "40.12"


@pytest.mark.parametrize("mqtt_topic_prefix", ["ft017th"])
def test__run_mock_measurements(mqtt_topic_prefix):
    # pylint: disable=too-many-arguments
    mqtt_client_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "wireless_sensor_mqtt._init_mqtt_client"
    ) as init_mqtt_client_mock, unittest.mock.patch(
        "time.sleep"
    ) as sleep_mock, unittest.mock.patch(
        "wireless_sensor_mqtt._publish_homeassistant_discovery_config"
    ) as hass_config_mock:
        init_mqtt_client_mock.return_value = mqtt_client_mock
        wireless_sensor_mqtt._run(
            mqtt_host="mqtt-broker.local",
            mqtt_port=1234,
            mqtt_disable_tls=True,
            mqtt_username=None,
            mqtt_password=None,
            mqtt_topic_prefix=mqtt_topic_prefix,
            homeassistant_discovery_prefix="homeassistant",
            homeassistant_node_id="ft017th-living-room",
            mock_measurements=True,
            unlock_spi_device=False,
        )
    assert init_mqtt_client_mock.call_count == 1
    assert hass_config_mock.call_count == 1
    assert all(c[0][0] == 8 for c in sleep_mock.call_args_list)
    publish_calls_args = mqtt_client_mock.publish.call_args_list
    assert len(publish_calls_args) == 2 * 3
    for call_args in publish_calls_args:
        assert not call_args[0]  # positional args
        assert set(call_args[1].keys()) == {"topic", "payload", "retain"}
        assert call_args[1]["retain"] is False
        assert float(call_args[1]["payload"]) > 0
    assert (
        publish_calls_args[0][1]["topic"]
        == publish_calls_args[2][1]["topic"]
        == publish_calls_args[4][1]["topic"]
        == mqtt_topic_prefix + "/temperature-degrees-celsius"
    )
    assert (
        publish_calls_args[1][1]["topic"]
        == publish_calls_args[3][1]["topic"]
        == publish_calls_args[5][1]["topic"]
        == mqtt_topic_prefix + "/relative-humidity-percent"
    )



# wireless-sensor-mqtt - MQTT client reporting measurements of FT017TH wireless thermo/hygrometers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import json
import logging
import pathlib
import random
import socket
import time
import typing

import paho.mqtt.client
import wireless_sensor

import wireless_sensor_mqtt._homeassistant

_MQTT_DEFAULT_PORT = 1883
_MQTT_DEFAULT_TLS_PORT = 8883
_MQTT_PUBLISH_TIMEOUT_SECONDS = 16
_MQTT_PUBLISH_STATUS_POLL_INTERVAL_SECONDS = 1
_MEASUREMENT_MOCKS_COUNT = 3
_MEASUREMENT_MOCKS_INTERVAL_SECONDS = 8

_LOGGER = logging.getLogger(__name__)


def _mqtt_on_connect(
    mqtt_client: paho.mqtt.client.Client,
    userdata: None,
    flags: typing.Dict,
    return_code: int,
) -> None:
    # pylint: disable=unused-argument; callback
    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.0/src/paho/mqtt/client.py#L441
    assert return_code == 0, return_code  # connection accepted
    mqtt_broker_host, mqtt_broker_port = mqtt_client.socket().getpeername()
    _LOGGER.debug("connected to MQTT broker %s:%d", mqtt_broker_host, mqtt_broker_port)


def _init_mqtt_client(
    # *, SyntaxError on python3.5
    host: str,
    port: int,
    username: typing.Optional[str],
    password: typing.Optional[str],
    disable_tls: bool,
) -> paho.mqtt.client.Client:
    # https://pypi.org/project/paho-mqtt/
    client = paho.mqtt.client.Client()
    client.on_connect = _mqtt_on_connect
    if not disable_tls:
        client.tls_set(ca_certs=None)  # enable tls trusting default system certs
    _LOGGER.info(
        "connecting to MQTT broker %s:%d (TLS %s)",
        host,
        port,
        "disabled" if disable_tls else "enabled",
    )
    if username:
        client.username_pw_set(username=username, password=password)
    elif password:
        raise ValueError("Missing MQTT username")
    client.connect(host=host, port=port)
    return client


def _mqtt_attempt_reconnect(client: paho.mqtt.client.Client) -> None:
    """
    > Calling connect() or reconnect() will cause the messages to be resent.
    > Use reinitialise() to reset a client to its original state.
    https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L530
    """
    try:
        return_code = client.reconnect()
        if return_code != paho.mqtt.client.MQTT_ERR_SUCCESS:
            raise RuntimeError(
                "Client.reconnect() returned expected return code {}".format(
                    return_code
                )
            )
    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L1805
    except (socket.error, OSError, RuntimeError):
        _LOGGER.error("failed to reconnect to MQTT broker", exc_info=True)


def _mock_measurement() -> wireless_sensor.Measurement:
    time.sleep(_MEASUREMENT_MOCKS_INTERVAL_SECONDS)
    return wireless_sensor.Measurement(
        decoding_timestamp=None,
        temperature_degrees_celsius=random.uniform(20.0, 30.0),
        relative_humidity=random.uniform(0.4, 0.6),
    )


def _mqtt_publish(
    *, client: paho.mqtt.client.Client, topic: str, payload: str, **kwargs
) -> None:
    _LOGGER.debug("publishing mqtt msg: topic=%s payload=%s", topic, payload)
    msg_info = client.publish(
        topic=topic, payload=payload, **kwargs
    )  # type: paho.mqtt.client.MQTTMessageInfo
    # MQTTMessageInfo.wait_for_publish() calls threading.Condition.wait() without timeout
    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L338
    poll_start_time = time.time()
    while (
        not msg_info.is_published()
        and (time.time() - poll_start_time) < _MQTT_PUBLISH_TIMEOUT_SECONDS
    ):
        if msg_info.rc == paho.mqtt.client.MQTT_ERR_NO_CONN:
            _mqtt_attempt_reconnect(client)
        time.sleep(_MQTT_PUBLISH_STATUS_POLL_INTERVAL_SECONDS)
    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L147
    if msg_info.rc != paho.mqtt.client.MQTT_ERR_SUCCESS:
        _LOGGER.error(
            "failed to publish on topic %s (return code %d)", topic, msg_info.rc
        )
    elif not msg_info.is_published():
        _LOGGER.warning(
            "reached timeout of %d seconds"
            " while waiting for MQTT message on topic %s to get published",
            _MQTT_PUBLISH_TIMEOUT_SECONDS,
            topic,
        )


def _publish_homeassistant_discovery_config(
    # *, SyntaxError on python3.5
    mqtt_client: paho.mqtt.client.Client,
    homeassistant_discovery_prefix: str,
    homeassistant_node_id: str,
    temperature_topic: str,
    humidity_topic: str,
) -> None:
    """
    https://www.home-assistant.io/docs/mqtt/discovery/
    """
    # topic format: <discovery_prefix>/<component>/[<node_id>/]<object_id>/config
    # https://www.home-assistant.io/docs/mqtt/discovery/
    # https://github.com/home-assistant/core/blob/0.117.5/homeassistant/components/mqtt/__init__.py#L274
    device_attrs = {
        # > voluptuous.error.MultipleInvalid: Device must have at least one id
        # > entifying value in 'identifiers' and/or 'connections'
        # > for dictionary value @ data['device']
        "identifiers": ["FT017TH/{}".format(homeassistant_node_id)],
        "model": "FT017TH",
    }
    for object_id, device_class, state_topic, unit, name_suffix in zip(
        ("temperature-degrees-celsius", "relative-humidity-percent"),
        # https://www.home-assistant.io/integrations/sensor/#device-class
        ("temperature", "humidity"),
        (temperature_topic, humidity_topic),
        ("°C", "%"),
        ("temperature", "relative humidity"),
    ):
        discovery_topic = "/".join(
            (
                homeassistant_discovery_prefix,
                "sensor",
                homeassistant_node_id,
                object_id,
                "config",
            )
        )
        unique_id = "/".join(
            (
                "fphammerle",
                "wireless-sensor-mqtt",
                "FT017TH",
                homeassistant_node_id,
                object_id,
            )
        )
        # https://www.home-assistant.io/integrations/sensor.mqtt/#configuration-variables
        # https://github.com/home-assistant/core/blob/0.117.5/homeassistant/components/mqtt/sensor.py#L50
        config = {
            "unique_id": unique_id,
            "device_class": device_class,
            # friendly_name & template for default entity_id
            "name": "{} {}".format(homeassistant_node_id, name_suffix),
            "state_topic": state_topic,
            "unit_of_measurement": unit,
            "expire_after": 60 * 10,  # seconds
            "device": device_attrs,
        }
        _mqtt_publish(
            client=mqtt_client,
            topic=discovery_topic,
            payload=json.dumps(config),
            retain=True,
        )


def _measurement_iter(
    mock_measurements: bool, unlock_spi_device: bool
) -> typing.Iterator[wireless_sensor.Measurement]:
    if mock_measurements:
        logging.warning("publishing %d mocked measurements", _MEASUREMENT_MOCKS_COUNT)
        return map(lambda _: _mock_measurement(), range(_MEASUREMENT_MOCKS_COUNT))
    return wireless_sensor.FT017TH(unlock_spi_device=unlock_spi_device).receive()


def _run(
    # *, SyntaxError on python3.5
    mqtt_host: str,
    mqtt_port: int,
    mqtt_disable_tls: bool,
    mqtt_username: typing.Optional[str],
    mqtt_password: typing.Optional[str],
    mqtt_topic_prefix: str,
    homeassistant_discovery_prefix: str,
    homeassistant_node_id: str,
    mock_measurements: bool,
    unlock_spi_device: bool,
) -> None:
    # pylint: disable=too-many-arguments
    # https://pypi.org/project/paho-mqtt/
    mqtt_client = _init_mqtt_client(
        host=mqtt_host,
        port=mqtt_port,
        disable_tls=mqtt_disable_tls,
        username=mqtt_username,
        password=mqtt_password,
    )
    temperature_topic = mqtt_topic_prefix + "/temperature-degrees-celsius"
    humidity_topic = mqtt_topic_prefix + "/relative-humidity-percent"
    _LOGGER.debug(
        "publishing measurements on topics %r and %r", temperature_topic, humidity_topic
    )
    homeassistant_discover_config_published = False
    for measurement in _measurement_iter(
        mock_measurements=mock_measurements, unlock_spi_device=unlock_spi_device
    ):
        _LOGGER.debug("received %s", measurement)
        if not homeassistant_discover_config_published:
            _publish_homeassistant_discovery_config(
                mqtt_client=mqtt_client,
                homeassistant_discovery_prefix=homeassistant_discovery_prefix,
                homeassistant_node_id=homeassistant_node_id,
                temperature_topic=temperature_topic,
                humidity_topic=humidity_topic,
            )
            homeassistant_discover_config_published = True
        _mqtt_publish(
            client=mqtt_client,
            topic=temperature_topic,
            payload="{:.02f}".format(measurement.temperature_degrees_celsius),
            retain=False,
        )
        _mqtt_publish(
            client=mqtt_client,
            topic=humidity_topic,
            # > unit_of_measurement: '%'
            # https://www.home-assistant.io/integrations/sensor.mqtt/#temperature-and-humidity-sensors
            payload="{:.02f}".format(measurement.relative_humidity * 100),
            retain=False,
        )


def _main() -> None:
    argparser = argparse.ArgumentParser(
        description="MQTT client reporting measurements of FT017TH wireless thermo/hygrometers",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argparser.add_argument("--mqtt-host", type=str, required=True)
    argparser.add_argument(
        "--mqtt-port",
        type=int,
        help="default {} ({} with --mqtt-disable-tls)".format(
            _MQTT_DEFAULT_TLS_PORT, _MQTT_DEFAULT_PORT
        ),
    )
    argparser.add_argument("--mqtt-username", type=str)
    argparser.add_argument("--mqtt-disable-tls", action="store_true")
    password_argument_group = argparser.add_mutually_exclusive_group()
    password_argument_group.add_argument("--mqtt-password", type=str)
    password_argument_group.add_argument(
        "--mqtt-password-file",
        type=pathlib.Path,
        metavar="PATH",
        dest="mqtt_password_path",
        help="stripping trailing newline",
    )
    argparser.add_argument(
        "--mqtt-topic-prefix",
        type=str,
        default="wireless-sensor/FT017TH",
        help=" ",  # show default
    )
    # https://www.home-assistant.io/docs/mqtt/discovery/#discovery_prefix
    argparser.add_argument(
        "--homeassistant-discovery-prefix", type=str, default="homeassistant", help=" "
    )
    argparser.add_argument(
        "--homeassistant-node-id",
        type=str,
        # pylint: disable=protected-access
        default="FT017TH",
        help=" ",
    )
    argparser.add_argument(
        "--mock-measurements",
        action="store_true",
        help="publish random values to test MQTT connection",
    )
    # https://github.com/fphammerle/wireless-sensor/blob/v0.3.0/wireless_sensor/_cli.py#L28
    argparser.add_argument(
        "--unlock-spi-device",
        action="store_true",
        help="Release flock from SPI device file after configuring the transceiver."
        " Useful if another process (infrequently) accesses the transceiver simultaneously.",
    )
    argparser.add_argument("--debug", action="store_true", help="increase verbosity")
    argparser.add_argument(
        "--debug-cc1101",
        action="store_true",
        help="increase verbosity of cc1101 library",
    )
    args = argparser.parse_args()
    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format="%(asctime)s:%(levelname)s:%(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )
    logging.getLogger("cc1101").setLevel(
        logging.DEBUG if args.debug_cc1101 else logging.INFO
    )
    if args.mqtt_port:
        mqtt_port = args.mqtt_port
    elif args.mqtt_disable_tls:
        mqtt_port = _MQTT_DEFAULT_PORT
    else:
        mqtt_port = _MQTT_DEFAULT_TLS_PORT
    if args.mqtt_password_path:
        # .read_text() replaces \r\n with \n
        mqtt_password = args.mqtt_password_path.read_bytes().decode()
        if mqtt_password.endswith("\r\n"):
            mqtt_password = mqtt_password[:-2]
        elif mqtt_password.endswith("\n"):
            mqtt_password = mqtt_password[:-1]
    else:
        mqtt_password = args.mqtt_password
    # pylint: disable=protected-access; false positive for validate_node_id
    if not wireless_sensor_mqtt._homeassistant.validate_node_id(
        args.homeassistant_node_id
    ):
        raise ValueError(
            "invalid home assistant node id {!r} (length >= 1, allowed characters: {})".format(
                args.homeassistant_node_id,
                # pylint: disable=protected-access; false positive
                wireless_sensor_mqtt._homeassistant.NODE_ID_ALLOWED_CHARS,
            )
            + "\nchange argument of --homeassistant-node-id"
        )
    _run(
        mqtt_host=args.mqtt_host,
        mqtt_port=mqtt_port,
        mqtt_disable_tls=args.mqtt_disable_tls,
        mqtt_username=args.mqtt_username,
        mqtt_password=mqtt_password,
        mqtt_topic_prefix=args.mqtt_topic_prefix,
        homeassistant_discovery_prefix=args.homeassistant_discovery_prefix,
        homeassistant_node_id=args.homeassistant_node_id,
        mock_measurements=args.mock_measurements,
        unlock_spi_device=args.unlock_spi_device,
    )



# wireless-sensor-mqtt - MQTT client reporting measurements of FT017TH wireless thermo/hygrometers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re

# https://www.home-assistant.io/docs/mqtt/discovery/#configuration-variables
NODE_ID_ALLOWED_CHARS = r"a-zA-Z0-9_-"


def validate_node_id(node_id: str) -> bool:
    return re.match(r"^[{}]+$".format(NODE_ID_ALLOWED_CHARS), node_id) is not None



# wireless-sensor-mqtt - MQTT client reporting measurements of FT017TH wireless thermo/hygrometers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import typing
import unittest.mock

import pytest

import wireless_sensor_mqtt

# pylint: disable=protected-access


@pytest.mark.parametrize(
    (
        "argv",
        "expected_mqtt_host",
        "expected_mqtt_port",
        "expected_mqtt_disable_tls",
        "expected_username",
        "expected_password",
        "expected_topic_prefix",
    ),
    [
        (
            ["", "--mqtt-host", "mqtt-broker.local"],
            "mqtt-broker.local",
            8883,
            False,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-disable-tls"],
            "mqtt-broker.local",
            1883,
            True,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-port", "8883"],
            "mqtt-broker.local",
            8883,
            False,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-port", "8884"],
            "mqtt-broker.local",
            8884,
            False,
            None,
            None,
            None,
        ),
        (
            [
                "",
                "--mqtt-host",
                "mqtt-broker.local",
                "--mqtt-port",
                "8884",
                "--mqtt-disable-tls",
            ],
            "mqtt-broker.local",
            8884,
            True,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-username", "me"],
            "mqtt-broker.local",
            8883,
            False,
            "me",
            None,
            None,
        ),
        (
            [
                "",
                "--mqtt-host",
                "mqtt-broker.local",
                "--mqtt-username",
                "me",
                "--mqtt-password",
                "secret",
            ],
            "mqtt-broker.local",
            8883,
            False,
            "me",
            "secret",
            None,
        ),
        (
            [
                "",
                "--mqtt-host",
                "mqtt-broker.local",
                "--mqtt-topic-prefix",
                "system/command",
            ],
            "mqtt-broker.local",
            8883,
            False,
            None,
            None,
            "system/command",
        ),
    ],
)
def test__main(
    argv,
    expected_mqtt_host,
    expected_mqtt_port,
    expected_mqtt_disable_tls,
    expected_username,
    expected_password,
    expected_topic_prefix: typing.Optional[str],
):
    # pylint: disable=too-many-arguments
    with unittest.mock.patch(
        "wireless_sensor_mqtt._run"
    ) as run_mock, unittest.mock.patch("sys.argv", argv):
        # pylint: disable=protected-access
        wireless_sensor_mqtt._main()
    run_mock.assert_called_once_with(
        mqtt_host=expected_mqtt_host,
        mqtt_port=expected_mqtt_port,
        mqtt_disable_tls=expected_mqtt_disable_tls,
        mqtt_username=expected_username,
        mqtt_password=expected_password,
        mqtt_topic_prefix=expected_topic_prefix or "wireless-sensor/FT017TH",
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_node_id="FT017TH",
        mock_measurements=False,
        unlock_spi_device=False,
    )


@pytest.mark.parametrize(
    ("password_file_content", "expected_password"),
    [
        ("secret", "secret"),
        ("secret space", "secret space"),
        ("secret   ", "secret   "),
        ("  secret ", "  secret "),
        ("secret\n", "secret"),
        ("secret\n\n", "secret\n"),
        ("secret\r\n", "secret"),
        ("secret\n\r\n", "secret\n"),
        ("你好\n", "你好"),
    ],
)
def test__main_password_file(tmpdir, password_file_content, expected_password):
    mqtt_password_path = tmpdir.join("mqtt-password")
    with mqtt_password_path.open("w") as mqtt_password_file:
        mqtt_password_file.write(password_file_content)
    with unittest.mock.patch(
        "wireless_sensor_mqtt._run"
    ) as run_mock, unittest.mock.patch(
        "sys.argv",
        [
            "",
            "--mqtt-host",
            "localhost",
            "--mqtt-username",
            "me",
            "--mqtt-password-file",
            str(mqtt_password_path),
        ],
    ):
        # pylint: disable=protected-access
        wireless_sensor_mqtt._main()
    run_mock.assert_called_once_with(
        mqtt_host="localhost",
        mqtt_port=8883,
        mqtt_disable_tls=False,
        mqtt_username="me",
        mqtt_password=expected_password,
        mqtt_topic_prefix="wireless-sensor/FT017TH",
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_node_id="FT017TH",
        mock_measurements=False,
        unlock_spi_device=False,
    )


def test__main_password_file_collision(capsys):
    with unittest.mock.patch(
        "sys.argv",
        [
            "",
            "--mqtt-host",
            "localhost",
            "--mqtt-username",
            "me",
            "--mqtt-password",
            "secret",
            "--mqtt-password-file",
            "/var/lib/secrets/mqtt/password",
        ],
    ):
        with pytest.raises(SystemExit):
            # pylint: disable=protected-access
            wireless_sensor_mqtt._main()
    out, err = capsys.readouterr()
    assert not out
    assert (
        "argument --mqtt-password-file: not allowed with argument --mqtt-password\n"
        in err
    )


@pytest.mark.parametrize(
    ("args", "discovery_prefix"),
    [
        ([], "homeassistant"),
        (["--homeassistant-discovery-prefix", "home/assistant"], "home/assistant"),
    ],
)
def test__main_homeassistant_discovery_prefix(args, discovery_prefix):
    with unittest.mock.patch(
        "wireless_sensor_mqtt._run"
    ) as run_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        wireless_sensor_mqtt._main()
    assert run_mock.call_count == 1
    assert run_mock.call_args[1]["homeassistant_discovery_prefix"] == discovery_prefix


@pytest.mark.parametrize(
    ("args", "node_id"),
    [
        ([], "FT017TH"),
        (["--homeassistant-node-id", "ft017th-living-room"], "ft017th-living-room"),
    ],
)
def test__main_homeassistant_node_id(args, node_id):
    with unittest.mock.patch(
        "wireless_sensor_mqtt._run"
    ) as run_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        wireless_sensor_mqtt._main()
    assert run_mock.call_count == 1
    assert run_mock.call_args[1]["homeassistant_node_id"] == node_id


@pytest.mark.parametrize(
    "args", [["--homeassistant-node-id", "no pe"], ["--homeassistant-node-id", ""]]
)
def test__main_homeassistant_node_id_invalid(args):
    with unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        with pytest.raises(ValueError):
            wireless_sensor_mqtt._main()


@pytest.mark.parametrize(
    ("additional_argv", "root_log_level"),
    [([], logging.INFO), (["--debug"], logging.DEBUG)],
)
def test__main_log_level(additional_argv, root_log_level):
    with unittest.mock.patch("wireless_sensor_mqtt._run"), unittest.mock.patch(
        "logging.basicConfig"
    ) as logging_basic_config_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + additional_argv
    ):
        wireless_sensor_mqtt._main()
    assert logging_basic_config_mock.call_count == 1
    assert logging_basic_config_mock.call_args[1]["level"] == root_log_level
    assert logging.getLogger("cc1101").getEffectiveLevel() == logging.INFO


@pytest.mark.parametrize(
    ("additional_argv", "log_level"),
    [([], logging.INFO), (["--debug-cc1101"], logging.DEBUG)],
)
def test__main_log_level_cc1101(additional_argv, log_level):
    with unittest.mock.patch("wireless_sensor_mqtt._run"), unittest.mock.patch(
        "logging.basicConfig"
    ) as logging_basic_config_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + additional_argv
    ):
        wireless_sensor_mqtt._main()
    assert logging_basic_config_mock.call_count == 1
    assert logging_basic_config_mock.call_args[1]["level"] == logging.INFO
    assert logging.getLogger("cc1101").getEffectiveLevel() == log_level


@pytest.mark.parametrize(
    ("additional_argv", "mock_measurements"),
    [([], False), (["--mock-measurements"], True)],
)
def test__main_mock_measurements(additional_argv, mock_measurements):
    with unittest.mock.patch(
        "wireless_sensor_mqtt._run"
    ) as main_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + additional_argv
    ):
        wireless_sensor_mqtt._main()
    assert main_mock.call_args[1]["mock_measurements"] == mock_measurements


@pytest.mark.parametrize(
    ("additional_argv", "unlock_spi_device"),
    [([], False), (["--unlock-spi-device"], True)],
)
def test__main_unlock_spi_device(additional_argv, unlock_spi_device):
    with unittest.mock.patch(
        "wireless_sensor_mqtt._run"
    ) as main_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + additional_argv
    ):
        wireless_sensor_mqtt._main()
    assert main_mock.call_args[1]["unlock_spi_device"] == unlock_spi_device



# wireless-sensor-mqtt - MQTT client reporting measurements of FT017TH wireless thermo/hygrometers
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import logging
import socket
import unittest.mock

import paho.mqtt.client
import pytest

import wireless_sensor_mqtt

# pylint: disable=protected-access


@pytest.mark.parametrize("host", ["mqtt-broker.local"])
@pytest.mark.parametrize("port", [8883, 1234])
def test__init_mqtt_client(caplog, host, port):
    # pylint: disable=too-many-locals,too-many-arguments
    caplog.set_level(logging.DEBUG)
    with unittest.mock.patch(
        "socket.create_connection"
    ) as create_socket_mock, unittest.mock.patch(
        "ssl.SSLContext.wrap_socket", autospec=True
    ) as ssl_wrap_socket_mock:
        ssl_wrap_socket_mock.return_value.send = len
        mqtt_client = wireless_sensor_mqtt._init_mqtt_client(
            host=host, port=port, username=None, password=None, disable_tls=False
        )
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message == (
        "connecting to MQTT broker {}:{} (TLS enabled)".format(host, port)
    )
    # correct remote?
    assert create_socket_mock.call_count == 1
    create_socket_args, _ = create_socket_mock.call_args
    assert create_socket_args[0] == (host, port)
    # ssl enabled?
    assert ssl_wrap_socket_mock.call_count == 1
    ssl_context = ssl_wrap_socket_mock.call_args[0][0]  # self
    assert ssl_context.check_hostname is True
    assert ssl_wrap_socket_mock.call_args[1]["server_hostname"] == host
    assert mqtt_client._tls_insecure is False
    # credentials
    assert mqtt_client._username is None
    assert mqtt_client._password is None
    # connect callback
    caplog.clear()
    mqtt_client.socket().getpeername.return_value = (host, port)
    # pylint: disable=not-callable; false positive
    mqtt_client.on_connect(mqtt_client, mqtt_client._userdata, {}, 0)
    assert caplog.records[0].levelno == logging.DEBUG
    assert caplog.records[0].message == "connected to MQTT broker {}:{}".format(
        host, port
    )


@pytest.mark.parametrize("host", ["mqtt-broker.local"])
@pytest.mark.parametrize("port", [1833])
@pytest.mark.parametrize("disable_tls", [True, False])
def test__init_mqtt_client_tls(caplog, host, port, disable_tls):
    caplog.set_level(logging.INFO)
    with unittest.mock.patch("paho.mqtt.client.Client") as mqtt_client_class:
        wireless_sensor_mqtt._init_mqtt_client(
            host=host, port=port, disable_tls=disable_tls, username=None, password=None
        )
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message == (
        "connecting to MQTT broker {}:{} (TLS {})".format(
            host, port, "disabled" if disable_tls else "enabled"
        )
    )
    if disable_tls:
        mqtt_client_class().tls_set.assert_not_called()
    else:
        mqtt_client_class().tls_set.assert_called_once_with(ca_certs=None)


@pytest.mark.parametrize("host", ["mqtt-broker.local"])
@pytest.mark.parametrize("port", [1833])
@pytest.mark.parametrize("username", ["me"])
@pytest.mark.parametrize("password", [None, "secret"])
def test__init_mqtt_client_authentication(host, port, username, password):
    with unittest.mock.patch("socket.create_connection"), unittest.mock.patch(
        "ssl.SSLContext.wrap_socket"
    ) as ssl_wrap_socket_mock:
        ssl_wrap_socket_mock.return_value.send = len
        mqtt_client = wireless_sensor_mqtt._init_mqtt_client(
            host=host,
            port=port,
            username=username,
            password=password,
            disable_tls=False,
        )
    assert mqtt_client._username.decode() == username
    if password:
        assert mqtt_client._password.decode() == password
    else:
        assert mqtt_client._password is None


@pytest.mark.parametrize("mqtt_host", ["mqtt-broker.local"])
@pytest.mark.parametrize("mqtt_port", [1833])
def test__init_mqtt_client_authentication_missing_username(mqtt_host, mqtt_port):
    with unittest.mock.patch("paho.mqtt.client.Client"):
        with pytest.raises(ValueError, match=r"^Missing MQTT username$"):
            wireless_sensor_mqtt._init_mqtt_client(
                host=mqtt_host,
                port=mqtt_port,
                disable_tls=False,
                username=None,
                password="secret",
            )


def test__mqtt_attempt_reconnect(caplog):
    client_mock = unittest.mock.MagicMock()
    client_mock.reconnect.return_value = paho.mqtt.client.MQTT_ERR_SUCCESS
    with caplog.at_level(logging.INFO):
        wireless_sensor_mqtt._mqtt_attempt_reconnect(client_mock)
    client_mock.reconnect.assert_called_once_with()
    assert caplog.record_tuples == []


def test__mqtt_attempt_reconnect_unexpected_return_code(caplog):
    client_mock = unittest.mock.MagicMock()
    client_mock.reconnect.return_value = 42
    with caplog.at_level(logging.ERROR):
        wireless_sensor_mqtt._mqtt_attempt_reconnect(client_mock)
    assert caplog.record_tuples == [
        ("wireless_sensor_mqtt", logging.ERROR, "failed to reconnect to MQTT broker")
    ]
    assert isinstance(caplog.records[0].exc_info[1], RuntimeError)


@pytest.mark.parametrize("exc_type", [socket.error, OSError])
def test__mqtt_attempt_reconnect_exception(caplog, exc_type):
    client_mock = unittest.mock.MagicMock()
    client_mock.reconnect.side_effect = exc_type("test")
    with caplog.at_level(logging.ERROR):
        wireless_sensor_mqtt._mqtt_attempt_reconnect(client_mock)
    assert caplog.record_tuples == [
        ("wireless_sensor_mqtt", logging.ERROR, "failed to reconnect to MQTT broker")
    ]
    assert isinstance(caplog.records[0].exc_info[1], exc_type)


def test__mqtt_publish(caplog):
    client_mock = unittest.mock.MagicMock()
    msg_info_mock = unittest.mock.MagicMock()
    msg_info_mock.rc = paho.mqtt.client.MQTT_ERR_SUCCESS
    msg_info_mock.is_published.side_effect = [False, False, False, True, True]
    client_mock.publish.return_value = msg_info_mock
    with unittest.mock.patch("time.sleep") as sleep_mock, caplog.at_level(
        logging.DEBUG
    ):
        wireless_sensor_mqtt._mqtt_publish(
            client=client_mock, topic="/some/topic", payload="test", retain=False
        )
    client_mock.publish.assert_called_once_with(
        topic="/some/topic", payload="test", retain=False
    )
    assert sleep_mock.call_args_list == [unittest.mock.call(1)] * 3
    assert caplog.record_tuples == [
        (
            "wireless_sensor_mqtt",
            logging.DEBUG,
            "publishing mqtt msg: topic=/some/topic payload=test",
        )
    ]


def test__mqtt_publish_timeout(caplog):
    client_mock = unittest.mock.MagicMock()
    msg_info_mock = unittest.mock.MagicMock()
    msg_info_mock.rc = paho.mqtt.client.MQTT_ERR_SUCCESS  # default
    msg_info_mock.is_published.return_value = False
    client_mock.publish.return_value = msg_info_mock
    with unittest.mock.patch(
        "time.time", side_effect=itertools.count()
    ), unittest.mock.patch("time.sleep") as sleep_mock, caplog.at_level(logging.DEBUG):
        wireless_sensor_mqtt._mqtt_publish(
            client=client_mock, topic="/some/topic", payload="test", retain=False
        )
    client_mock.publish.assert_called_once_with(
        topic="/some/topic", payload="test", retain=False
    )
    assert sleep_mock.call_args_list == [unittest.mock.call(1)] * 15
    assert caplog.record_tuples == [
        (
            "wireless_sensor_mqtt",
            logging.DEBUG,
            "publishing mqtt msg: topic=/some/topic payload=test",
        ),
        (
            "wireless_sensor_mqtt",
            logging.WARNING,
            "reached timeout of 16 seconds while waiting for MQTT message on topic"
            " /some/topic to get published",
        ),
    ]


def test__mqtt_publish_fail(caplog):
    client_mock = unittest.mock.MagicMock()
    msg_info_mock = unittest.mock.MagicMock()
    msg_info_mock.rc = 42
    msg_info_mock.is_published.return_value = False
    client_mock.publish.return_value = msg_info_mock
    with unittest.mock.patch(
        "time.time", side_effect=itertools.count()
    ), unittest.mock.patch("time.sleep"), caplog.at_level(logging.ERROR):
        wireless_sensor_mqtt._mqtt_publish(
            client=client_mock, topic="/some/topic", payload="test", retain=False
        )
    assert caplog.record_tuples == [
        (
            "wireless_sensor_mqtt",
            logging.ERROR,
            "failed to publish on topic /some/topic (return code 42)",
        )
    ]


def test__mqtt_publish_reconnect():
    client_mock = unittest.mock.MagicMock()
    msg_info_mock = unittest.mock.MagicMock()
    msg_info_mock.rc = paho.mqtt.client.MQTT_ERR_NO_CONN
    msg_info_mock.is_published.return_value = False
    client_mock.publish.return_value = msg_info_mock
    with unittest.mock.patch(
        "wireless_sensor_mqtt._mqtt_attempt_reconnect"
    ) as reconnect_mock, unittest.mock.patch(
        "time.time", side_effect=itertools.count()
    ), unittest.mock.patch(
        "time.sleep"
    ):
        wireless_sensor_mqtt._mqtt_publish(
            client=client_mock, topic="/some/topic", payload="test", retain=False
        )
    client_mock.publish.assert_called_once_with(
        topic="/some/topic", payload="test", retain=False
    )
    assert reconnect_mock.call_count > 0
